/**************************************************************************
* iptrk - ip tracker
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia.
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* real time tracking routines
*
**************************************************************************/
#include <curses.h>
#include "defs.h"
#include "lib.p"
#include "iptrk.p"

PUBLIC addr_t         filter_src, filter_dst;
PUBLIC int            stream_factor = 2,
                      rate_gram_avg;

PRIVATE int           rate_gram;
PRIVATE U_32          rate;
PRIVATE voldg_t       rate_g;
PRIVATE volp_t        rate_p;
PRIVATE int           mode;
PRIVATE iptrk_queue_t iptrk_queue;
PRIVATE U_16          mport_s, mport_d;

/**************************************************************************
* pull entry off device and append to bt's
**************************************************************************/
PRIVATE int poll_device(void)
{
  char      str[STR_LEN + 1],
            str2[STR_LEN + 1];
  int       pack_dir = 0, n;
  our_net_t *s_net = NULL,
            *d_net = NULL;

  timer = time(NULL);
  time_st = localtime(&timer);

  for(n = 0; n != sizeof(iptrk_queue_t);)
  {
    n += read(iptrk_fd, &iptrk_queue + n, sizeof(iptrk_queue_t) - n);
    if(n == 0)
      return 0;
  }

  if(boot.debug_bt != NULL)
    if(!find_debug_addr(iptrk_queue.src, iptrk_queue.dst))
      return 1;

  get_our_net(iptrk_queue.src, &s_net);
  get_our_net(iptrk_queue.dst, &d_net);

 /*
  * cancel packet when-
  *   - both src and dst are in our nets
  *   - src & dst in different nets
  *   - dev == dev associated with dst IP
  */
  if(boot.cancel_gram)
    if(s_net != NULL && d_net != NULL)
      if(strcmp(s_net->dev, d_net->dev))
        if(strcmp(iptrk_queue.dev, d_net->dev))
        {
          if(mode == 0)
            display_stream_grams(
              iptrk_queue.src,
              iptrk_queue.dst,
              iptrk_queue.vol, 
              iptrk_queue.prot,
              iptrk_queue.port_s,
              iptrk_queue.port_d,
              -1);
          return 1;
        }

 /*
  * if streaming mode, print entry (before struct are clipped !!)
  */
  rate_gram++;
  rate += iptrk_queue.vol;

  if     (s_net != NULL && d_net != NULL)   /*DIR_u*/
  {
    pack_dir = DIR_u;
    rate_g.u += iptrk_queue.vol;
  }
  else if(s_net == NULL && d_net != NULL)   /*DIR_I*/
  {
    pack_dir = DIR_I;
    rate_g.I += iptrk_queue.vol;
  }
  else if(s_net != NULL && d_net == NULL)   /*DIR_O*/
  {
    pack_dir = DIR_O;
    rate_g.O += iptrk_queue.vol;
  }
  else if(s_net == NULL && d_net == NULL)   /*DIR_m*/
  {
    pack_dir = DIR_m;
    rate_g.m += iptrk_queue.vol;
  }

  switch(iptrk_queue.prot)
  {
    case IPPROTO_TCP:
      rate_p.tcp += iptrk_queue.vol;
      break;

    case IPPROTO_UDP:
      rate_p.udp += iptrk_queue.vol;
      break;

    default:
      rate_p.oth += iptrk_queue.vol;
      break;
  }

  if(mode == 0)
    display_stream_grams(
      iptrk_queue.src,
      iptrk_queue.dst,
      iptrk_queue.vol,
      iptrk_queue.prot,
      iptrk_queue.port_s,
      iptrk_queue.port_d,
      pack_dir);

 /*
  * do any clipping
  */
  if(watch_bt_size_lim_f)   /*watch db limit reached, clip external*/
  {                         /*addresses to all 1's*/
    if(s_net == NULL)
      iptrk_queue.src = 0xffffffff;
    if(d_net == NULL)
      iptrk_queue.dst = 0xffffffff;
  }
  else
  {
    if(watch_bt_size > boot.db_limit)  /*we're being attacked.. clip db*/
    {
      watch_bt_size_lim_f = 1;
      sprintf(str, " watch db limit reached ! (limit %ld). clipping entries",
        boot.db_limit);
      log_m('I', NULL, str);
      strcpy(str2, "\n" URL_NAME "\n");
      email_operator(str, str2);
    }
  }
  mport_s = mod_port(iptrk_queue.port_s);
  mport_d = mod_port(iptrk_queue.port_d);

 /*
  * update btree
  */
  update_ip_entry(
    &watch_bt,
    iptrk_queue.src,
    iptrk_queue.dst,
    iptrk_queue.vol,
    iptrk_queue.prot,
    mport_s,
    mport_d);

  return 1;
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void print_general(void)
{
  char str[STR_LEN + 1];

  attrset(COLOR_PAIR(COLOUR_WHITE) | A_BOLD);
  move(LINES - 3, 2);
  sprintf(str, "Total: %8lu Bps", rate);
  addstr(str);

  move(LINES - 3, 27);
  sprintf(str, "Pkt/Usg: %8d", rate_gram);
  addstr(str);

  attrset(COLOR_PAIR(COLOUR_CYAN) | A_BOLD);
  move(LINES - 2, 6);
  sprintf(str, "I: %8llu Bps", rate_g.I);
  addstr(str);

  move(LINES - 1, 6);
  sprintf(str, "O: %8llu Bps", rate_g.O);
  addstr(str);

  move(LINES - 2, 33);
  sprintf(str, "u: %8llu Bps", rate_g.u);
  addstr(str);

  move(LINES - 1, 33);
  sprintf(str, "m: %8llu Bps", rate_g.m);
  addstr(str);

  attrset(COLOR_PAIR(COLOUR_GREEN) | A_BOLD);

  move(LINES - 3, 58);
  sprintf(str, "TCP: %8llu Bps", rate_p.tcp);
  addstr(str);

  move(LINES - 2, 58);
  sprintf(str, "UDP: %8llu Bps", rate_p.udp);
  addstr(str);

  move(LINES - 1, 58);
  sprintf(str, "OTH: %8llu Bps", rate_p.oth);
  addstr(str);

  refresh();
}

/**************************************************************************
* if queue empty, then display periodical info
**************************************************************************/
PRIVATE void display_periodical(void)
{
  static time_t old_s_update;

  rate_gram_avg = (rate_gram_avg + rate_gram) >> 1;

  if(timer - old_s_update == 0)
    return;

  if(mode == 1)
    display_sorted_grams(
      iptrk_queue.src,
      iptrk_queue.dst,
      iptrk_queue.vol,
      iptrk_queue.prot,
      iptrk_queue.port_s,
      iptrk_queue.port_d,
      0);
  if(mode == 2)
    display_sorted_grams(
      iptrk_queue.src,
      iptrk_queue.dst,
      iptrk_queue.vol,
      iptrk_queue.prot,
      iptrk_queue.port_s,
      iptrk_queue.port_d,
      1);

  print_general();
  rate = 0;
  rate_gram = 0;
  memset(&rate_g, 0, sizeof(voldg_t));
  memset(&rate_p, 0, sizeof(volp_t));
  old_s_update = timer;
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void display_addr(U_32 a)
{
  struct in_addr in;
  char           str[STR_LEN + 1];

  if(a != 0xffffffff)
  {
    in.s_addr = htonl(a);   
    sprintf(str, "%-18s ", inet_ntoa(in)); /* xxx.xxx.xxx.xxx/xx */
    addstr(str);
  }   
  else
    addstr("all                ");
}

PRIVATE void display_filters(void)
{
  attrset(COLOR_PAIR(COLOUR_MAGENTA) | A_BOLD);

  move(LINES - 5, 1);
  addstr("IP SRC: ");
  display_addr(filter_src.addr);
  display_addr(filter_src.mask);

  move(LINES - 4, 1);
  addstr("IP DST: ");
  display_addr(filter_dst.addr);
  display_addr(filter_dst.mask);

  refresh();
}

PRIVATE void clear_filters(void)
{
  filter_src.addr = 0xffffffff;
  filter_dst.addr = 0xffffffff;
  filter_src.mask = 0xffffffff;
  filter_dst.mask = 0xffffffff;

  display_filters();
}

PRIVATE void edit_filter(char *str, int y, int x)
{
  attrset(COLOR_PAIR(COLOUR_EDIT) | A_REVERSE);
  mvhline(y, x, ' ', 18);
  getstr_rt_init(stdscr, y, x, 18, str);

  for(;;)
  {
    if(poll_device())
      continue;

    display_periodical();

    attrset(COLOR_PAIR(COLOUR_EDIT) | A_REVERSE);
    if(getstr_rt())
      break;

    usleep(boot.wait_delay);
  }
}

PRIVATE void edit_addr_filter(addr_t *f, int y, int x)
{
  char           str[STR_LEN + 1], *s;
  int            i;
  struct in_addr in;

  edit_filter(str, y, x);
  if(str[0] == 0)
  {
    f->addr = f->mask = 0xffffffff;
    display_filters();
    return;
  }

  i = 0;
  for(s = str; *s != 0; s++)
    if(*s == '.')
      i++;

  if(i != 3)
  {
    display_filters();
    return;
  }

  s = strchr(str, '/');
  if(s != NULL)
  {
    *s = 0;
    s++;
    i = atoi(s);
    if(i > 0 && i <= 32)
    {
      f->mask = 0;
      for(; i > 0; --i)
      {
        f->mask >>= 1;
        f->mask |= 0x80000000;
      }
    }
  }

  inet_aton(str, &in);
  f->addr = ntohl(in.s_addr);
  f->addr &= f->mask;
  display_filters();
}

PRIVATE void edit_mask_filter(addr_t *f, int y, int x)
{
  char           str[STR_LEN + 1];
  struct in_addr in;

  edit_filter(str, y, x);
  inet_aton(str, &in);
  f->mask = ntohl(in.s_addr);
  f->addr &= f->mask;
  display_filters();
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void draw_stream_factor(void)
{
  char str[STR_LEN + 1];

  attrset(COLOR_PAIR(COLOUR_TOPBAR2) | A_BOLD | A_REVERSE);
  move(0, 40);
  sprintf(str, "(sf: %d) ", stream_factor);
  addstr(str);
  attrset(COLOR_PAIR(COLOUR_NORMAL));
  refresh();
}

/***************************************************************************
*
***************************************************************************/
PUBLIC void gram_mode_stream(void)
{
  switch(mode)
  {
    case 0:
      mvwaddstr(win_gram_b, 0, 10, "(streaming )");
      break;

    case 1:
      mvwaddstr(win_gram_b, 0, 10, "(in  sorted)");
      break;

    case 2:
      mvwaddstr(win_gram_b, 0, 10, "(out sorted)");
      break;
  }

  wrefresh(win_gram_b);
}

/**************************************************************************
*
**************************************************************************/
PUBLIC void draw_screen(int bar_f)
{
  clear();
  refresh();

  if(bar_f)
    attrset(COLOR_PAIR(COLOUR_TOPBARO) | A_BOLD | A_REVERSE);
  else
    attrset(COLOR_PAIR(COLOUR_TOPBAR) | A_BOLD | A_REVERSE);

  mvhline(0, 0, ' ', 80);
  move(0, 0);
  addstr(" iptrk R" REVISION);
 
  if(boot.debug_bt != NULL)
    addstr(" [DEBUG]");
    
  move(0, 49);
  addstr("Copyright (C) 2003 Jason Nunn");

  print_general();
  display_filters();
  draw_stream_factor();

  box(win_gram_b, ACS_VLINE, ACS_HLINE);
  box(win_log_b, ACS_VLINE, ACS_HLINE);
  gram_mode_stream();
  wrefresh(win_log_b);
}

/**************************************************************************
*
**************************************************************************/
PUBLIC void rt_track(void)
{
  int ch;

  old_watch_time = timer;
  log_m('I', NULL, " polling device");

  nodelay(stdscr, 1);
  clear_filters();
  draw_screen(0);
  for(;;)
  {

   /*
    * poll device.. display streaming data while data in buffer
    */
    if(poll_device())
      continue;

   /*
    * do periodical & maintenance functions.
    */
    display_periodical();
    do_maintenance();

   /*
    * lastly, do input
    */
    ch = getch();
    if(ch != ERR)
    {
      if(ch == CTRL('W'))
      {
        log_m('I', NULL, " quit key hit..");
        break;
      }
      switch(ch)
      {
        case KEY_CR:
          mode++;
          if(mode > 2)
            mode = 0;
          gram_mode_stream();
          werase(win_gram);
          wrefresh(win_gram);
          break;

        case CTRL('L'):
          draw_screen(0);
          break;

        case '-':
          stream_factor--;
          if(stream_factor < 0)
            stream_factor = 0;
          draw_stream_factor();
          break;

        case '+':
          stream_factor++;
          if(stream_factor > SF_MAX)
            stream_factor = SF_MAX;
          draw_stream_factor();
          break;

        case 'i':
          edit_addr_filter(&filter_src, LINES - 5, 9);
          break;

        case 'o':
          edit_addr_filter(&filter_dst, LINES - 4, 9);
          break;

        case CTRL('I'):
          edit_mask_filter(&filter_src, LINES - 5, 28);
          break;

        case CTRL('O'):
          edit_mask_filter(&filter_dst, LINES - 4, 28);
          break;

        case 'u':
          clear_filters();
          break;

        case ' ':
          wattrset(win_gram, COLOR_PAIR(COLOUR_WHITE));
          wattron(win_gram, A_BOLD);
          waddstr(win_gram, "\n***** BREAK *****");
          wrefresh(win_gram);
          break;
      }
    }
    usleep(boot.wait_delay);
  }
  nodelay(stdscr, 0);
}
